"""
Methods for graphing volatility data

"""
from collections import Counter
import copy
import os
import warnings
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.graph_objects as go
import scipy as sp
from mpl_toolkits.mplot3d import Axes3D # pylint: disable=unused-import
from plotly.offline import plot
from scipy.interpolate import griddata
# pylint: disable=invalid-name

class Graph():
    """
    Methods for graphing volatility data

    """
    @classmethod
    def line_graph(cls, params, tables):
        """
        Displays a linegraph of each option maturity plotted by strike
        and implied vol

        Parameters
        ----------
        voltype : Str
            Whether to use 'bid', 'mid', 'ask' or 'last' price. The
            default is 'last'.
        save_image : Bool
            Whether to save a copy of the image as a png file. The default is
            False
        image_folder : Str
            Location to save any images. The default is 'images'
        image_dpi : Int
            Resolution to save images. The default is 50.

        Returns
        -------
        Linegraph.

        """

        # Create a sorted list of the different number of option
        # expiries
        dates = sorted(list(set(tables['imp_vol_data']['Expiry'])))

        # Create a sorted list of the different number of option time
        # to maturity
        tenors = sorted(list(set(tables['imp_vol_data']['TTM'])))

        # Combine these in a dictionary
        tenor_date_dict = dict(zip(dates, tenors))

        plt.style.use('seaborn-darkgrid')
        plt.rcParams.update(params['mpl_line_params'])
        fig_size = (12, 9)

        # Create figure, axis objects
        fig, ax = plt.subplots(figsize=fig_size)

        # Create values that scale fonts with fig_size
        ax_font_scale = int(round(fig_size[0] * 1.5))
        title_font_scale = int(round(fig_size[0] * 2))

        # Set fontsize of axis ticks
        ax.tick_params(axis='both', which='major', labelsize=ax_font_scale)

        # For each expiry date
        for exp_date, tenor in tenor_date_dict.items():

            # Plot the specified voltype against strike
            ax.plot(
                np.array(tables['imp_vol_data'][
                    tables['imp_vol_data']['TTM']==tenor]['Strike']),
                np.array(tables['imp_vol_data'][
                    tables['imp_vol_data']['TTM']==tenor][str(
                        params['vols_dict'][str(params['voltype'])])] * 100),
                label=str(exp_date)+' Expiry')

        plt.grid(True)

        # Label axes
        ax.set_xlabel('Strike', fontsize=ax_font_scale)
        ax.set_ylabel('Implied Volatility %', fontsize=ax_font_scale)

        # Set legend title and font sizes
        ax.legend(title="Option Expiry",
                  fontsize=ax_font_scale*0.6,
                  title_fontsize=ax_font_scale*0.8)

        # Specify title with ticker label, voltype and date and shift
        # away from chart
        st = fig.suptitle(
            str(params['ticker_label'])
            +' Implied Volatility '
            +str(params['voltype'].title())
            +' Price '
            +str(params['start_date']),
            fontsize=title_font_scale,
            fontweight=0,
            color='black',
            style='italic',
            y=1.02)
        st.set_y(0.95)
        fig.subplots_adjust(top=0.9)

        # Display graph
        plt.show()

        if params['save_image']:
            # save the image as a png file
            fig = cls._image_save(params=params, fig=fig)

        return params, tables


    @classmethod
    def scatter_3d(cls, params, tables):
        """
        Displays a 3D scatter plot of each option implied vol against
        strike and maturity

        Parameters
        ----------
        voltype : Str
            Whether to use 'bid', 'mid', 'ask' or 'last' price. The
            default is 'last'.
        azim : Float
            L-R view angle for 3D graphs. The default is -50.
        elev : Float
            Elevation view angle for 3D graphs. The default is 20.
        fig_size : Tuple
            3D graph size
        save_image : Bool
            Whether to save a copy of the image as a png file. The default is
            False
        image_folder : Str
            Location to save any images. The default is 'images'
        image_dpi : Int
            Resolution to save images. The default is 50.

        Returns
        -------
        3D Scatter plot.

        """

        # Create figure and axis objects and format
        fig, ax = cls._graph_format(params=params)

        # Create copy of data
        tables['data_3D'] = copy.deepcopy(tables['imp_vol_data'])

        # Filter out any zero prices
        tables['data_3D'] = tables['data_3D'][tables['data_3D'][str(
            params['prices_dict'][str(params['voltype'])])] != 0]

        # Specify the 3 axis values
        x = tables['data_3D']['Strike']
        y = tables['data_3D']['TTM'] * 365
        z = tables['data_3D'][str(params['vols_dict'][str(
            params['voltype'])])] * 100

        # Display scatter, specifying colour to vary with z-axis and use
        # colormap 'viridis'
        ax.scatter3D(x, y, z, c=z, cmap='viridis')

        if params['save_image']:
            # save the image as a png file
            fig = cls._image_save(params=params, fig=fig)

        return params, tables


    @classmethod
    def surface_3d(cls, params, tables):
        """
        Displays a 3D surface plot of the implied vol surface against
        strike and maturity

        Parameters
        ----------
        surfacetype : Str
            The type of 3D surface to display from 'trisurf', 'mesh',
            'spline', 'interactive_mesh' and 'interactive_spline'.
            The default is 'mesh'.
        smoothing : Bool
            Whether to apply polynomial smoothing. The default is False.
        scatter : Bool
            Whether to plot scatter points on 3D mesh grid. The default
            is False.
        voltype : Str
            Whether to use 'bid', 'mid', 'ask' or 'last' price. The
            default is 'last'.
        order : Int
            Polynomial order used in numpy polyfit function. The
            default is 3.
        spacegrain : Int
            Number of points in each axis linspace argument for 3D
            graphs. The default
            is 100.
        azim : Float
            L-R view angle for 3D graphs. The default is -50.
        elev : Float
            Elevation view angle for 3D graphs. The default is 20.
        fig_size : Tuple
            3D graph size
        rbffunc : Str
            Radial basis function used in interpolation chosen from
            'multiquadric', 'inverse', 'gaussian', 'linear', 'cubic',
            'quintic', 'thin_plate'. The default is 'thin_plate'
        colorscale : Str
            Colors used in plotly interactive graph. The default is
            'BlueRed'
        opacity : Float
            opacity of 3D interactive graph
        surf : Bool
            Plot surface in interactive graph
        notebook : Bool
            Whether interactive graph is run in Jupyter notebook or IDE.
            The default is False.
        save_image : Bool
            Whether to save a copy of the image as a png file. The default is
            False
        image_folder : Str
            Location to save any images. The default is 'images'
        image_dpi : Int
            Resolution to save images. The default is 50.

        Returns
        -------
        3D surface plot.

        """

        # Suppress mpl user warning about data containing nan values
        warnings.filterwarnings(
            "ignore", category=UserWarning, message='Z contains NaN values. '\
                'This may result in rendering artifacts.')

        # If smoothing is set to False
        if params['smoothing'] is False:

            # Create copy of implied vol data
            tables['data_3D'] = copy.deepcopy(tables['imp_vol_data'])

            # Filter out any zero prices
            tables['data_3D'] = tables['data_3D'][tables['data_3D'][str(
                params['prices_dict'][str(params['voltype'])])] != 0]

            # Set 'graph vol' to be the specified voltype
            tables['data_3D']['Graph Vol'] = tables['data_3D'][str(
                params['vols_dict'][str(params['voltype'])])]

        # Otherwise, if smoothing is set to True
        else:

            # Apply the smoothing function to the specified voltype
            params, tables = cls._smooth(params=params, tables=tables)

            # Create copy of implied vol data
            tables['data_3D'] = copy.deepcopy(tables['imp_vol_data_smoothed'])

            # Filter out any zero prices
            tables['data_3D'] = tables['data_3D'][tables['data_3D'][str(
                params['prices_dict'][str(params['voltype'])])] != 0]

            # Set 'graph vol' to be the smoothed vol
            tables['data_3D']['Graph Vol'] = tables['data_3D']['Smoothed Vol']

        # Specify the 3 axis values
        params['x'] = tables['data_3D']['Strike']
        params['y'] = tables['data_3D']['TTM'] * 365
        params['z'] = tables['data_3D']['Graph Vol'] * 100

        if params['surfacetype'] == 'trisurf':
            fig = cls._trisurf_graph(params=params)

        elif params['surfacetype'] == 'mesh':
            fig = cls._mesh_graph(params=params)

        elif params['surfacetype'] == 'spline':
            fig = cls._spline_graph(params=params, tables=tables)

        elif params['surfacetype'] in ['interactive_mesh',
                                       'interactive_spline']:
            fig = cls._interactive_graph(params=params, tables=tables)

        else:
            print("Enter a valid surfacetype from 'trisurf', 'mesh', "\
                  "'spline', 'interactive_mesh', 'interactive_spline'")

        if params['save_image']:
            # save the image as a png file
            fig = cls._image_save(params=params, fig=fig)

        # Set warnings back to default
        warnings.filterwarnings("default", category=UserWarning)

        return params, tables


    @classmethod
    def _trisurf_graph(cls, params):

        # Create figure and axis objects and format
        fig, ax = cls._graph_format(params=params)

        # Display triangular surface plot, using colormap 'viridis'
        ax.plot_trisurf(params['x'],
                        params['y'],
                        params['z'],
                        cmap='viridis',
                        edgecolor='none')

        return fig


    @classmethod
    def _mesh_graph(cls, params):

        # Create arrays across x and y-axes of equally spaced points
        # from min to max values
        x1, y1 = np.meshgrid(
            np.linspace(min(params['x']),
                        max(params['x']),
                        int(params['spacegrain'])),
            np.linspace(min(params['y']),
                        max(params['y']),
                        int(params['spacegrain'])))

        # Map the z-axis with the scipy griddata method, applying
        # cubic spline interpolation
        z1 = griddata(np.array([params['x'],params['y']]).T,
                      np.array(params['z']),
                      (x1,y1),
                      method='cubic')

        # Create figure and axis objects and format
        fig, ax = cls._graph_format(params=params)

        # Plot the surface
        ax.plot_surface(x1, y1, z1)

        # Apply contour lines
        ax.contour(x1, y1, z1)

        plt.show()

        return fig


    @classmethod
    def _spline_graph(cls, params, tables):

        # Create arrays across x and y-axes of equally spaced points
        # from min to max values
        x1 = np.linspace(min(params['x']),
                         max(params['x']),
                         int(params['spacegrain']))
        y1 = np.linspace(min(params['y']),
                         max(params['y']),
                         int(params['spacegrain']))
        x2, y2 = np.meshgrid(x1, y1, indexing='xy')

        # Initialize the z-axis as an array of zero values
        z2 = np.zeros((params['x'].size, params['z'].size))

        # Apply scipy interpolate radial basis function, choosing
        # the rbffunc parameter
        spline = sp.interpolate.Rbf(
            params['x'],
            params['y'],
            params['z'],
            function=params['rbffunc'],
            smooth=5,
            epsilon=5)

        # Populate z-axis array using this function
        z2 = spline(x2, y2)

        # Create figure and axis objects and format
        fig, ax = cls._graph_format(params=params)

        # Plot the surface
        ax.plot_wireframe(x2, y2, z2)
        ax.plot_surface(x2, y2, z2, alpha=0.2)

        # If scatter is True, overlay the surface with the
        # unsmoothed scatter points
        if params['scatter']:
            params['z'] = tables['data_3D'][str(
                params['vols_dict'][str(params['voltype'])])] * 100
            ax.scatter3D(params['x'], params['y'], params['z'], c='r')

        return fig

    @classmethod
    def _interactive_graph(cls, params, tables):

        params = cls._set_contours(params=params, tables=tables)

        # Specify the 3 axis values
        params['x'] = tables['data_3D']['TTM'] * 365
        params['y'] = tables['data_3D']['Strike']
        params['z'] = tables['data_3D']['Graph Vol'] * 100

        # Create arrays across x and y-axes of equally spaced
        # points from min to max values
        x1 = np.linspace(
            params['x'].min(),
            params['x'].max(),
            int(params['spacegrain'])
            )
        y1 = np.linspace(
            params['y'].min(),
            params['y'].max(),
            int(params['spacegrain'])
            )
        params['x2'], params['y2'] = np.meshgrid(x1, y1, indexing='xy')

        # If surfacetype is 'interactive_mesh', map the z-axis with
        # the scipy griddata method, applying cubic spline
        # interpolation
        if params['surfacetype'] == 'interactive_mesh':
            params['z2'] = griddata((params['x'], params['y']),
                                    params['z'],
                                    (params['x2'], params['y2']),
                                    method='cubic')

        # If surfacetype is 'interactive_spline', apply scipy
        # interpolate radial basis function, choosing the rbffunc
        # parameter
        if params['surfacetype'] == 'interactive_spline':
            params['z2'] = np.zeros((params['x'].size, params['z'].size))
            spline = sp.interpolate.Rbf(
                params['x'],
                params['y'],
                params['z'],
                function=params['rbffunc'],
                smooth=5,
                epsilon=5)
            params['z2'] = spline(params['x2'], params['y2'])

        # Initialize Figure object
        if params['scatter']:
            fig = cls._int_scatter(params=params, tables=tables)

        # Plot just the surface
        else:
            fig = cls._int_surf(params=params)

        # Format the chart layout
        fig = cls._int_layout(params=params, fig=fig)

        # If running within a Jupyter notebook, plot graph inline
        if params['notebook'] is True:
            fig.show()

        # Otherwise create a new HTML window to display
        else:
            plot(fig, auto_open=True)

        return fig


    @staticmethod
    def _set_contours(params, tables):

        # Set the range of x, y and z contours and interval
        params['contour_x_start'] = 0
        params['contour_x_stop'] = 2 * 360
        params['contour_x_size'] = params['contour_x_stop'] / 18
        params['contour_y_start'] = tables['data_3D']['Strike'].min()
        params['contour_y_stop'] = tables['data_3D']['Strike'].max()

        # Vary the strike interval based on spot level
        if ((tables['data_3D']['Strike'].max()
             - tables['data_3D']['Strike'].min()) > 2000):
            params['contour_y_size'] = 200
        elif ((tables['data_3D']['Strike'].max()
               - tables['data_3D']['Strike'].min()) > 1000):
            params['contour_y_size'] = 100
        elif ((tables['data_3D']['Strike'].max()
               - tables['data_3D']['Strike'].min()) > 250):
            params['contour_y_size'] = 50
        elif ((tables['data_3D']['Strike'].max()
               - tables['data_3D']['Strike'].min()) > 50):
            params['contour_y_size'] = 10
        else:
            params['contour_y_size'] = 5

        # Set z contours
        params['contour_z_start'] = 0
        params['contour_z_stop'] = 100
        params['contour_z_size'] = 5

        return params


    @staticmethod
    def _int_scatter(params, tables):

        # Set z to raw data points
        params['z'] = (tables['data_3D'][str(
            params['vols_dict'][str(params['voltype'])])] * 100)

        # Create figure object with fitted surface and scatter
        # points
        if params['surf']:
            fig = go.Figure(
                data=[go.Surface(
                    x=params['x2'],
                    y=params['y2'],
                    z=params['z2'],

                    # Specify the colors to be used
                    colorscale=params['colorscale'],

                    # Define the contours
                    contours = {
                        "x": {
                            "show": True,
                            "start":params['contour_x_start'],
                            "end":params['contour_x_stop'],
                            "size":params['contour_x_size'],
                            "color":"white"
                            },
                        "y": {
                            "show": True,
                            "start":params['contour_y_start'],
                            "end":params['contour_y_stop'],
                            "size":params['contour_y_size'],
                            "color":"white"
                            },
                        "z": {
                            "show": True,
                            "start":params['contour_z_start'],
                            "end": params['contour_z_stop'],
                            "size": params['contour_z_size']
                            }
                        },

                    # Set the surface opacity
                    opacity=params['opacity']),

                # Plot scatter of unsmoothed data
                go.Scatter3d(
                    x=params['x'],
                    y=params['y'],
                    z=params['z'],
                    mode='markers',
                    # Set size, color and opacity of each data point
                    marker=dict(
                        size=2,
                        color='red',
                        opacity=0.9
                        )
                    )
                ])

        # Plot just the scatter points
        else:
            fig = go.Figure(data=[
                # Plot scatter of unsmoothed data
                go.Scatter3d(
                    x=params['x'],
                    y=params['y'],
                    z=params['z'],
                    mode='markers',
                    # Set size, color and opacity of each data point
                    marker=dict(
                        size=2,
                        color='red',
                        opacity=0.9
                        )
                    )
                ])

        return fig


    @staticmethod
    def _int_surf(params):

        # Create figure object with fitted surface
        fig = go.Figure(
            data=[go.Surface(
                x=params['x2'],
                y=params['y2'],
                z=params['z2'],

                # Specify the colors to be used
                colorscale=params['colorscale'],

                # Define the contours
                contours = {
                    "x": {
                        "show": True,
                        "start":params['contour_x_start'],
                        "end":params['contour_x_stop'],
                        "size":params['contour_x_size'],
                        "color":"white"
                        },
                    "y": {
                        "show": True,
                        "start":params['contour_y_start'],
                        "end":params['contour_y_stop'],
                        "size":params['contour_y_size'],
                        "color":"white"
                        },
                    "z": {
                        "show": True,
                        "start":params['contour_z_start'],
                        "end":params['contour_z_stop'],
                        "size":params['contour_z_size']
                        }
                    },

                   # Set the surface opacity
                   opacity=params['opacity'])])

        return fig


    @staticmethod
    def _int_layout(params, fig):

        # Set initial camera angle
        params['camera'] = dict(
            eye=dict(x=2, y=1, z=1)
        )

        # Set Time To Expiration to increase left to right
        fig.update_scenes(xaxis_autorange="reversed")
        fig.update_layout(
            scene = dict(
                xaxis = dict(
                    backgroundcolor="rgb(200, 200, 230)",
                    gridcolor="white",
                    showbackground=True,
                    zerolinecolor="white",),
                yaxis = dict(
                    backgroundcolor="rgb(230, 200,230)",
                    gridcolor="white",
                    showbackground=True,
                    zerolinecolor="white"),
                zaxis = dict(
                    backgroundcolor="rgb(230, 230,200)",
                    gridcolor="white",
                    showbackground=True,
                    zerolinecolor="white",),
                # Label axes
                xaxis_title='Time to Expiration (Days)',
                yaxis_title='Strike',
                zaxis_title='Implied Volatility %',),
            # Specify title with ticker label, voltype
            # and date
            title={'text':(str(params['ticker_label'])
                           +' Implied Volatility '
                           +str(params['voltype'].title())
                           +' Price '
                           +str(params['start_date'])),
                   'y':0.9,
                   'x':0.5,
                   'xanchor':'center',
                   'yanchor':'top',
                   'font':dict(size=20,
                               color="black")},
            autosize=False,
            width=800,
            height=800,
            margin=dict(l=65, r=50, b=65, t=90),
            scene_camera=params['camera'])

        return fig


    @staticmethod
    def _graph_format(params):

        # Update chart parameters
        plt.rcParams.update(params['mpl_3D_params'])

        # Create fig object
        fig = plt.figure(figsize=params['fig_size'])

        # Create axes object
        ax = fig.add_subplot(
            111,
            projection='3d',
            azim=params['azim'],
            elev=params['elev'])

        # Set background color to white
        ax.set_facecolor('w')

        # Create values that scale fonts with fig_size
        ax_font_scale = int(round(params['fig_size'][0] * 1.1))
        title_font_scale = int(round(params['fig_size'][0] * 1.5))

        # Tint the axis panes, RGB values from 0-1 and alpha denoting
        # color intensity
        ax.w_xaxis.set_pane_color((0.9, 0.8, 0.9, 0.8))
        ax.w_yaxis.set_pane_color((0.8, 0.8, 0.9, 0.8))
        ax.w_zaxis.set_pane_color((0.9, 0.9, 0.8, 0.8))

        # Set z-axis to left hand side
        ax.zaxis._axinfo['juggled'] = (1, 2, 0) # pylint: disable=protected-access

        # Set fontsize of axis ticks
        ax.tick_params(axis='both', which='major', labelsize=ax_font_scale)

        # Label axes
        ax.set_xlabel('Strike', fontsize=ax_font_scale,
                      labelpad=ax_font_scale*1.2)
        ax.set_ylabel('Time to Expiration (Days)', fontsize=ax_font_scale,
                      labelpad=ax_font_scale*1.2)
        ax.set_zlabel('Implied Volatility %', fontsize=ax_font_scale,
                      labelpad=ax_font_scale*1.2)

        # Specify title with ticker label, voltype and date
        st = fig.suptitle(str(params['ticker_label'])
                          +' Implied Volatility '
                          +str(params['voltype'].title())
                          +' Price '
                          +str(params['start_date']),
                          fontsize=title_font_scale,
                          fontweight=0,
                          color='black',
                          style='italic',
                          y=1.02)

        st.set_y(0.95)
        fig.subplots_adjust(top=1)

        return fig, ax


    @classmethod
    def _smooth(cls, params, tables):
        """
        Create a column of smoothed implied vols

        Parameters
        ----------
        order : Int
            Polynomial order used in numpy polyfit function. The
            default is 3.
        voltype : Str
            Whether to use 'bid', 'mid', 'ask' or 'last' price. The
            default is 'last'.
        smoothopt : Int
            Minimum number of options to fit curve to. The default
            is 6.

        Returns
        -------
        DataFrame
            DataFrame of Option prices.

        """

        # Create a dictionary of the number of options for each
        # maturity
        mat_dict = dict(Counter(tables['imp_vol_data']['Days']))

        # Create a sorted list of the different number of days to
        # maturity
        maturities = sorted(list(set(tables['imp_vol_data']['Days'])))

        # Create a sorted list of the different number of strikes
        strikes_full = sorted(list(set((tables['imp_vol_data'][
            'Strike'].astype(float)))))

        # create copy of implied vol data
        tables['imp_vol_data_smoothed'] = copy.deepcopy(tables['imp_vol_data'])

        for ttm, count in mat_dict.items():

            # if there are less than smoothopt (default is 6) options
            # for a given maturity
            if count < params['smoothopt']:

                # remove that maturity from the maturities list
                maturities.remove(ttm)

                # and remove that maturity from the implied vol
                # DataFrame
                tables['imp_vol_data_smoothed'] = tables[
                    'imp_vol_data_smoothed'][
                        tables['imp_vol_data_smoothed']['Days'] != ttm]

        # Create empty DataFrame with the full range of strikes as
        # index
        tables['smooth_surf'] = pd.DataFrame(index=strikes_full)

        # going through the maturity list (in reverse so the columns
        # created are in increasing order)
        for maturity in reversed(maturities):

            # Extract the strikes for this maturity
            strikes = tables['imp_vol_data'][tables['imp_vol_data'][
                'Days']==maturity]['Strike']

            # And the vols (specifying the voltype)
            vols = tables['imp_vol_data'][tables['imp_vol_data'][
                'Days']==maturity][str(
                    params['vols_dict'][str(params['voltype'])])]

            # Fit a polynomial to this data
            curve_fit = np.polyfit(strikes, vols, params['order'])
            p = np.poly1d(curve_fit)

            # Create empty list to store smoothed implied vols
            iv_new = []

            # For each strike
            for strike in strikes_full:

                # Add the smoothed value to the iv_new list
                iv_new.append(p(strike))

            # Append this list as a new column in the smooth_surf
            # DataFrame
            tables['smooth_surf'].insert(0, str(maturity), iv_new)

        # Apply the _vol_map function to add smoothed vol column to
        # DataFrame
        tables['imp_vol_data_smoothed'] = (
            tables['imp_vol_data_smoothed'].apply(
                lambda x: cls._vol_map(x, tables), axis=1))

        return params, tables


    @staticmethod
    def _vol_map(row, tables):
        """
        Map value calculated in smooth surface DataFrame to
        'Smoothed Vol' column.

        Parameters
        ----------
        row : Array
            Each row in the DataFrame.

        Returns
        -------
        row : Array
            Each row in the DataFrame.

        """
        row['Smoothed Vol'] = (
            tables['smooth_surf'].loc[row['Strike'], str(row['Days'])])

        return row


    @staticmethod
    def _image_save(params, fig):

        # Create image folder if it does not already exist
        if not os.path.exists(params['image_folder']):
            os.makedirs(params['image_folder'])

        # save the image as a png file
        plt.savefig('{}/{}{}.png'.format(
            params['image_folder'], params['ticker_label'],
            params['start_date']), dpi=params['image_dpi'])

        return fig



import volatility
vis = volatility.Volatility()

#%%
vis_dict = vis.__dict__

#%%
import datetime as dt
from datetime import timedelta
from pandas.tseries.offsets import BDay

start_date_as_dt = (dt.datetime.today() - timedelta(days=1) - BDay(1)).date()
start_date = str(start_date_as_dt)

#%%
params = vis.params
url = 'https://finance.yahoo.com/quote/'+params['ticker']\
            +'/options?p='+params['ticker']

#%%
import requests
# Define the stock root webpage
url = 'https://finance.yahoo.com/quote/'+params['ticker']\
    +'/options?p='+params['ticker']

#%%
# Create a requests object to extract data from the url
params['requestslink'] = requests.get(url)

#%%
# Collect the text fromthis object
html_doc = params['requestslink'].text

#%%
# Use Beautiful Soup to parse this
soup = BeautifulSoup(html_doc, features="lxml")

# Create a list of all the option dates
option_dates = [a.get_text() for a in soup.find_all('option')]

# Convert this list from string to datetimes
dates_list = [dt.datetime.strptime(date, "%B %d, %Y").date() for date
              in option_dates]

# Convert back to strings in the required format
str_dates = [date_obj.strftime('%Y-%m-%d') for date_obj in dates_list]

# Create a list of all the unix dates used in the url for each
# of these dates
option_pages = [a.attrs['value'] for a in soup.find_all('option')]

# Combine the dates and unixdates in a dictionary
optodict = dict(zip(str_dates, option_pages))

# Create an empty dictionary
url_dict = {}

# For each date and unixdate in the first dictionary
for date_val, page in optodict.items():

    # Create an entry with the date as key and the url plus
    # unix date as value
    url_dict[date_val] = str(
        'https://finance.yahoo.com/quote/'
        +params['ticker']+'/options?date='+page)

#%%
# https://github.com/JECSand/yahoofinancials/blob/9cc2bd1faa7380f6f451f7de9b5d228160627fd3/yahoofinancials/__init__.py#L70

def _get_api_data(self, api_url, tries=0):
    urlopener = UrlOpener()
    response = urlopener.open(api_url)
    if response.getcode() == 200:
        res_content = response.read()
        response.close()
        if sys.version_info < (3, 0):
            return loads(res_content)
        return loads(res_content.decode('utf-8'))
    else:
        if tries < 5:
            time.sleep(random.randrange(10, 20))
            tries += 1
            return self._get_api_data(api_url, tries)
        else:
            return None

#%%
# https://docs.python.org/3/library/urllib.request.html#module-urllib.request
#try:
#from urllib import FancyURLopener
#except:
from urllib.request import FancyURLopener

#%%

# Class used to open urls for financial data
class UrlOpener(FancyURLopener):
    version = 'w3m/0.5.3+git20180125'

#%%
urlopener = UrlOpener()

#%%
response = urlopener.open(url)

#%%
response_content = response.read()

#%%
from bs4 import BeautifulSoup
soup = BeautifulSoup(response_content, "html.parser")

#%%
print(soup.prettify())

#%%
print(soup.get_text())

#%%
htmldoc = response.text

#%%
# Use Beautiful Soup to parse this
soup = BeautifulSoup(response_content, features="lxml")

#%%

# Create a list of all the option dates
option_dates = [a.get_text() for a in soup.find_all('option')]

#%%
# Convert this list from string to datetimes
dates_list = [dt.datetime.strptime(date, "%B %d, %Y").date() for date
              in option_dates]

#%%
# Convert back to strings in the required format
str_dates = [date_obj.strftime('%Y-%m-%d') for date_obj in dates_list]

#%%
# Create a list of all the unix dates used in the url for each
# of these dates
option_pages = [a.attrs['value'] for a in soup.find_all('option')]

#%%
# Combine the dates and unixdates in a dictionary
optodict = dict(zip(str_dates, option_pages))

#%%
# Create an empty dictionary
url_dict = {}

# For each date and unixdate in the first dictionary
for date_val, page in optodict.items():

    # Create an entry with the date as key and the url plus
    # unix date as value
    url_dict[date_val] = str(
        'https://finance.yahoo.com/quote/'
        +params['ticker']+'/options?date='+page)

#%%
import volatility
vis = volatility.Volatility(ticker='AAPL')

#%%
vis_dict = vis.__dict__

#%%
# Create an empty dictionary
option_dict = {}
params['url_except_dict'] = {}

# each url needs to have an option expiry date associated with
# it in the url dict
for input_date, url in params['url_dict'].items():

    # requests function downloads the data
    web = requests.get(url).content

    # wait between each query so as not to overload server
    time.sleep(params['wait'])

    # if data exists
    try:
        # read html data into a DataFrame
        option_frame = pd.read_html(web)

        # Add this DataFrame to the default dictionary, named
        # with the expiry date it refers to
        params['option_dict'][input_date] = option_frame

    # otherwise collect dictionary of exceptions
    except KeyError:
        params['url_except_dict'][input_date] = url


#%%
params = vis.params

#%%
url_dict = params['url_dict']

#%%
urldates = list(url_dict.keys())
urllist = list(url_dict.values())

#%%

urldate1 = urldates[0]
url1 = urllist[0]

#%%
import pandas as pd
web = requests.get(url).content
option_frame = pd.read_html(web)

#%%
web2 = urlopener.open(url1)
web2_content = web2.read()
option_frame2 = pd.read_html(web2_content)

#%%
#for input_date, url in params['url_dict'].items():

urldate1 = urldates[0]
url1 = urllist[0]

# UrlOpener function downloads the data
urlopener = UrlOpener()
weburl = urlopener.open(url1)
web = weburl.read()

# if data exists

# read html data into a DataFrame
option_frame = pd.read_html(web)

#%%

# Add this DataFrame to the default dictionary, named
# with the expiry date it refers to
params['option_dict'][input_date]['calls'] = option_frame[0]
params['option_dict'][input_date]['puts'] = option_frame[1]

#%%
# otherwise collect dictionary of exceptions
except KeyError:
    params['url_except_dict'][input_date] = url

#%%
import copy
vis_bak = copy.deepcopy(vis)

#%%
vis = copy.deepcopy(vis_bak)

#%%
from market_data import Data
vis.params, vis.tables = Data.transform(params=vis.params, tables=vis.tables)

#%%
# Create Days to Maturity column
days = (vis.tables['data']['TTM']*365).astype(int)

#%%
import numpy as np
# Create a column of the Trade Date represented in unixtime
unixtime2 = (vis.tables['data']['Last Trade Date'].view(np.int64) // 10**9)

#%%
vis.params, vis.tables = Data.combine(params=vis.params, tables=vis.tables)

#%%
visdict = vis.__dict__

#%%
from lxml import html
data = urlopener.open(url1)
data_content = data.read()
tree = html.fromstring(data_content)

#%%
priceparse = tree.xpath(
    '//span[@class="Trsdu(0.3s) Fw(b) Fz(36px) Mb(-4px) D(ib)"]/text()')

spot = float([str(p) for p in priceparse][0].replace(',',''))

#%%
urlopener = UrlOpener()
response = urlopener.open(url)

# Collect the text from this object
params['html_doc'] = response.read()

#%%
input_data = copy.deepcopy(vis.tables['data'])

#%%
# For each put strike price
for strike in params['put_strikes']:

    # Assign an option name of ticker plus strike
    opt_name = params['ticker_label']+'_'+str(strike)

    # store the implied vol results for that strike in the
    # option dictionary
    opt_dict[opt_name] = cls._imp_vol_apply(
        params=params, input_data=input_data, K=strike, option='put')

    # store the implied vol results for that strike in the
    # option list
    params['opt_list'].append(opt_dict[opt_name])

    print('Put option: ', opt_name)

#%%
params = vis.params

#%%
put0 = params['put_strikes'][0]

#%%
import warnings
strike=80
aapl80 = _imp_vol_apply(params=params, input_data=input_data, K=strike, option='put')

#%%
def _imp_vol_apply(params, input_data, K, option):

    # Filter data by strike and option type
    input_data = (input_data[(input_data['Strike'] == K)
            & (input_data['Option Type'] == option)])

    # Apply implied vol method to each row
    input_data = input_data.apply(
        lambda x: _imp_vol_by_row(x, params, K, option), axis=1)

    return input_data

#%%
from optionmodels.models import ImpliedVol
def _imp_vol_by_row(row, params, K, option):

    # Suppress runtime warnings caused by bad vol data
    warnings.filterwarnings("ignore", category=RuntimeWarning)

    # For the chosen implied vol method and its method name
    for flag, func_name in params['method_dict'].items():

        # select the method from the dictionary
        if params['method'] == flag:

            # for each of the prices: bid, mid, ask, last
            for input_row, output_row in params['row_dict'].items():

                try:
                    # populate the column using the chosen implied
                    # vol method (using getattr() to select
                    # dynamically)
                    # check if n/a value is returned and print error
                    # message if so
                    output = getattr(ImpliedVol, func_name)(
                        S=params['S'], K=K, T=row['TTM'],
                        r=params['r'], q=params['q'],
                        cm=row[input_row], epsilon=params['epsilon'],
                        option=option, timing=False)

                    output = float(output)
                    row[output_row] = output

                except KeyError:
                    print("Error with option: Strike="+str(K)+
                              " TTM="+str(round(row['TTM'], 3))+
                              " vol="+str(row[input_row])+
                              " option="+option)

    # Return warnings to default setting
    warnings.filterwarnings("default", category=RuntimeWarning)

    return row

#%%
def _imp_vol_by_row(row, params, K, option):

    # Suppress runtime warnings caused by bad vol data
    warnings.filterwarnings("ignore", category=RuntimeWarning)

    func_name = params['method_dict'][params['method']]

    # for each of the prices: bid, mid, ask, last
    for input_row, output_row in params['row_dict'].items():

        #try:
            # populate the column using the chosen implied
            # vol method (using getattr() to select
            # dynamically)
            # check if n/a value is returned and print error
            # message if so
        output = getattr(ImpliedVol, func_name)(
            S=params['spot'], K=K, T=row['TTM'],
            r=params['r'], q=params['q'],
            cm=row[input_row], epsilon=params['epsilon'],
            option=option, timing=False)

        output = float(output)
        row[output_row] = output

        #except KeyError:
        #    print("Error with option: Strike="+str(K)+
        #              " TTM="+str(round(row['TTM'], 3))+
        #              " vol="+str(row[input_row])+
        #              " option="+option)

    # Return warnings to default setting
    warnings.filterwarnings("default", category=RuntimeWarning)

    return row

#%%
func_name = params['method_dict'][params['method']]

#%%
from graph import Graph
params, tables = Graph.line_graph(params=vis.params, tables=vis.tables)

#%%
params, tables = Graph.scatter_3d(params=vis.params, tables=vis.tables)

#%%
        if params['surfacetype'] == 'trisurf':
            fig = cls._trisurf_graph(params=params)

        elif params['surfacetype'] == 'mesh':
            fig = cls._mesh_graph(params=params)

        elif params['surfacetype'] == 'spline':
            fig = cls._spline_graph(params=params, tables=tables)

        elif params['surfacetype'] in ['interactive_mesh',
                                       'interactive_spline']:

#%%
params['surfacetype'] = 'trisurf'
params, tables = Graph.surface_3d(params=vis.params, tables=vis.tables)

#%%
params['surfacetype'] = 'mesh'
params, tables = Graph.surface_3d(params=vis.params, tables=vis.tables)

#%%
params['surfacetype'] = 'spline'
params['scatter'] = True
params, tables = Graph.surface_3d(params=vis.params, tables=vis.tables)

#%%
params['surfacetype'] = 'interactive_mesh'
#params['scatter'] = True
params, tables = Graph.surface_3d(params=vis.params, tables=vis.tables)

#%%
params['surfacetype'] = 'interactive_spline'
#params['scatter'] = True
params, tables = Graph.surface_3d(params=vis.params, tables=vis.tables)

#%%
from volatility import Volatility
spx = Volatility()

#%%
spxdict = spx.__dict__

#%%
from market_data import Data
raw_web_data = Data._extract_web_data(params=spx.params)

#%%
import copy
spxcopy = copy.deepcopy(spx)

#%%
spx.params['raw_web_data'] = raw_web_data
spx.params = Data._read_web_data(params=spx.params)

#%%
import pandas as pd
# Create an empty DataFrame
spx.tables['full_data'] = pd.DataFrame()

# Make a list of all the dates of the DataFrames just stored
# in the default dictionary
spx.params['date_list'] = list(spx.params['option_dict'].keys())
spx.params, spx.tables = Data._process_options(params=spx.params, tables=spx.tables)

#%%
from market_data import Data
spx.params, spx.tables = Data.extractoptions(params=spx.params, tables=spx.tables)

#%%
spx.params, spx.tables = Data.transform(params=spx.params, tables=spx.tables)

#%%
spx.params, spx.tables = Data.combine(params=spx.params, tables=spx.tables)

#%%
from volatility import Volatility
aapl = Volatility(ticker='AAPL')

#%%
aapldict = aapl.__dict__

#%%
aapl.visualize(graphtype='line')

#%%
aapl.visualize(graphtype='scatter')

#%%
aapl.visualize(graphtype='surface', surfacetype='trisurf')

#%%
aapl.visualize(graphtype='surface', surfacetype='mesh')

#%%
aapl.visualize(graphtype='surface', surfacetype='spline')

#%%
aapl.visualize(graphtype='surface', surfacetype='interactive_mesh')

#%%
aapl.visualize(graphtype='surface', surfacetype='interactive_spline')

#%%
from volatility import Volatility
tsla = Volatility(ticker='TSLA')

#%%
tsladict = tsla.__dict__

#%%
params = tsla.params

#%%
from market_data import Data
aapl.params, aapl.tables = Data._process_options(aapl.params, aapl.tables)

#%%
aapl.params, aapl.tables = Data.transform(params=aapl.params, tables=aapl.tables)

#%%
aapl.params, aapl.tables = Data.combine(params=aapl.params, tables=aapl.tables)

#%%
from volvisualizer.volatility import Volatility
amzn = Volatility(ticker='AMZN')

#%%
amzn.linegraph()

#%%
amzn.scatter()

#%%
amzn.surface(surfacetype='spline')

#%%
from volvisualizer.volatility import Volatility
pfe = Volatility(ticker='PFE')

#%%
pfe.linegraph()

#%%
pfe.scatter()

#%%
pfe.surface(surfacetype='spline')

#%%
pfedict = pfe.__dict__

#%%
from volvisualizer.volatility import Volatility

#%%
nvda = Volatility(ticker='NVDA')

#%%
nvdadict = nvda.__dict__

#%%
params = nvda.params

#%%
params = msft.params

#%%
roundspot = (
    round(params['spot'] / params['divisor']) * params['divisor'])

#%%
divisor = 1

#%%
put_min = (round(params['spot']
                 * params['strike_limits'][0]
                 / params['divisor'])
           * params['divisor'])

#%%
put_strikes = list(
    range(put_min, roundspot, params['divisor']))

#%%
import numpy as np
pt2 = list(np.linspace(put_min, roundspot, 45))

#%%
divisor = 2.5
((roundspot - put_min) / divisor) + 1

#%%
divisor = params['divisor']

#%%
import numpy as np
pt3 = np.linspace(put_min, roundspot, int((roundspot - put_min) / divisor) + 1)

#%%
pt4 = np.linspace(roundspot, call_max, int((call_max - roundspot) / divisor) + 1)

#%%
divisor = 1.25

#%%
pfe = Volatility(ticker='PFE')

#%%
from volatility import Volatility
msft = Volatility(ticker='MSFT')

#%%
nvdadict = nvda.__dict__

#%%
strikes = set(nvda.tables['data']['Strike'])

#%%
strikes2 = set(msft.tables['data']['Strike'])

#%%
div2_5 = {x for x in strikes if x%(2.5)==0}

#%%
params = msft.params

#%%
divisor = 10

#%%
put_min = (round(params['spot']
                 * params['strike_limits'][0]
                 / params['divisor'])
           * params['divisor'])

call_max = (round(params['spot']
                  * params['strike_limits'][1]
                  / params['divisor'])
            * params['divisor'])

#%%
div10 = {x for x in strikes if (x%(10)==0 and put_min < x < call_max)}

#%%
avail_strikes = {}
for div in [0.5, 1, 1.25, 2.5, 5, 10]:
    avail_strikes[div] = len(
        {x for x in strikes if (x%(div)==0 and put_min < x < call_max)})

#%%
as2 = {}
for div in [0.5, 1, 1.25, 2.5, 5, 10]:
    as2[div] = len(
        {x for x in strikes2 if (x%(div)==0 and put_min < x < call_max)})

#%%
nvda = Volatility(ticker='NVDA')

#%%
max(avail_strikes, key=avail_strikes.get)

#%%
max(as2, key=as2.get)

#%%
import operator
max(as2.items(), key=operator.itemgetter(1))[0]

#%%

strikes = set(nvda.tables['data']['Strike'])

put_min = (round(params['spot']
                 * params['strike_limits'][0]
                 / divisor)
           * divisor)

call_max = (round(params['spot']
                  * params['strike_limits'][1]
                  / divisor)
            * divisor)


avail_strikes = {}
for div in [0.5, 1, 1.25, 2.5, 5, 10]:
    avail_strikes[div] = len(
        {x for x in strikes if (x%(div)==0 and put_min < x < call_max)})


as2 = avail_strikes

max([divisor for max_strike_count in [
    max(avail_strikes.values())] for divisor, strike_count in avail_strikes.items()
    if strike_count == max_strike_count])

#%%
msft.visualize(graphtype='line')

#%%
from market_data import Data
params = msft.params
msft.params, msft.tables = Data.combine(params=msft.params, tables=msft.tables)

#%%
msftdict = msft.__dict__

#%%
nvda.visualize(graphtype='surface', smooth=True)

#%%
from volatility import Volatility
spx = Volatility(ticker='^SPX', wait=0.5, q=0.013)

#%%
spxdict=spx.__dict__

#%%
spx.visualize(graphtype='line')

#%%
spx.visualize(graphtype='surface', smooth=True)

#%%
gld = Volatility(ticker='GLD')
glddict= gld.__dict__

#%%
gld.visualize(graphtype='line')

#%%
# Create an empty list
date_list = []


#%%
import datetime as dt
params = spx.params

# For each date in the url_dict
for key in params['url_dict'].keys():

    # Format that string as a datetime object
    key_date = dt.datetime.strptime(key, "%Y-%m-%d")

    # Store the year and month as a tuple in date_list
    date_list.append((key_date.year, key_date.month))


#%%
# Create a sorted list from the unique dates in date_list
sorted_dates = sorted(list(set(date_list)))


#%%
# Create an empty list
days_list = []

#%%
import calendar
# Create a calendar object
c = calendar.Calendar(firstweekday=calendar.SATURDAY)


#%%
# For each tuple of year, month in sorted_dates
for tup in sorted_dates:

    # Create a list of lists of days in that month
    monthcal = c.monthdatescalendar(tup[0], tup[1])

    # Extract the date corresponding to the 3rd Friday
    expiry = monthcal[2][-1]

    # Calculate the number of days until that expiry
    ttm = (expiry - dt.date.today()).days

    # Append this to the days_list
    days_list.append(ttm)

#%%
tables = spx.tables
full_day_list = set(tables['full_data']['Days'])

#%%
# For each unique number of days to expiry
for days_to_expiry in set(tables['data']['Days']):

    # if the expiry is not in the list of monthly expiries
    if days_to_expiry not in days_list:

        # Remove that expiry from the DataFrame
        tables['data'] = (
            tables['data'][tables['data']['Days']
                           != days_to_expiry])


#%%
from market_data import Data
par2, tab2 = Data.transform(params=spx.params, tables=spx.tables)

#%%
tab2['data']['AnnTTM'] = tab2['data']['TTM']*365

#%%
(tab2['data']['AnnTTM']).dtype

#%%
import math
tab2['data']['Days2'] = math.floor(tab2['data']['TTM']*365)

#%%
import pandas as pd
from datetime import date
ttm = (pd.to_datetime(tables['data']['Expiry'])
    - pd.to_datetime(date.today())) / (pd.Timedelta(days=1) * 365)

#%%
# Create Days to Maturity column
days = (tables['data']['TTM']*365)#.astype(int)

#%%
days_set = sorted(list(set(days)))

#%%
days_list = []
exps = []
for tup in sorted_dates:

    # Create a list of lists of days in that month
    monthcal = c.monthdatescalendar(tup[0], tup[1])

    # Extract the date corresponding to the 3rd Friday
    expiry = monthcal[2][-1]

    exps.append(expiry)

    # Calculate the number of days until that expiry
    ttm = (expiry - dt.date.today()).days

    # Append this to the days_list
    days_list.append(ttm)


#%%
tab2['data']['ttm1'] = (pd.to_datetime(tab2['data']['Expiry'])
    - pd.to_datetime(date.today())) / (pd.Timedelta(days=1) * 365)

#%%
tab2['data']['today'] = pd.to_datetime(dt.date.today())

#%%
import numpy as np
tab2['data']['ttm2'] = np.array([0.0]*len(tab2['data']))
for row in tab2['data']:
    tab2['data']['ttm2'][row] = (pd.to_datetime(tab2['data']['Expiry'][row]) - tab2['data']['today'][row]).days


#%%
# Create Days to Maturity column
tab2['data']['dt1'] = (tab2['data']['TTM']*365).astype(int)
tab2['data']['dt2'] = np.round(tab2['data']['TTM']*365, 0)
tab2['data']['dt3'] = tab2['data']['TTM']*365

#%%
spx.params, spx.tables = Data.transform(params=spx.params, tables=spx.tables)

#%%
spx.params, spx.tables = Data.combine(params=spx.params, tables=spx.tables)

#%%
amc = Volatility(ticker='AMC')

#%%
amc.visualize(graphtype='line')

#%%
nvda = Volatility(ticker='NVDA', wait=0.5)

#%%
nvda.visualize(graphtype='surface', surfacetype='spline', scatter=True, smoothing=True)

#%%
nvda.visualize(graphtype='surface',surfacetype='interactive_spline', smoothing=True, notebook=False, colorscale='Blues', scatter=True, opacity=0.8)


"""
Extract option prices from Yahoo Finance and visualize volatility

"""

import copy
from volvisualizer.graph import Graph
from volvisualizer.market_data import Data
from volvisualizer.utils import Utils
from volvisualizer.volatility_params import vol_params_dict

class Volatility():
    """
    Extract option prices from Yahoo Finance and visualize volatility

    """
    def __init__(self, **kwargs):
        """
        Extract the URL for each of the listed option on Yahoo Finance
        for the given ticker. Extract option data from each URL.

        Filter / transform the data and calculate implied volatilities for
        specified put and call strikes.

        Parameters
        ----------
        start_date : Str
            Date from when to include prices (some of the options
            won't have traded for days / weeks and therefore will
            have stale prices).
        ticker : Str
            The ticker identifier used by Yahoo for the chosen stock. The
            default is '^SPX'.
        ticker_label : TYPE, optional
            DESCRIPTION. The default is None.
        wait : Int
            Number of seconds to wait between each url query
        lastmins : Int, Optional
            Restrict to trades within number of minutes since last
            trade time recorded
        mindays : Int, Optional
            Restrict to options greater than certain option expiry
        minopts : Int, Optional
            Restrict to minimum number of options to include that
            option expiry
        volume : Int, Optional
            Restrict to minimum Volume
        openint : Int, Optional
            Restrict to minimum Open Interest
        monthlies : Bool
            Restrict expiries to only 3rd Friday of the month. Default
            is False.
        spot : Float
            Underlying reference level.
        put_strikes : List
            Range of put strikes to calculate implied volatility for.
        call_strikes : List
            Range of call strikes to calculate implied volatility for.
        strike_limits : Tuple
            min and max strikes to use expressed as a decimal
            percentage. The default is (0.5, 2.0).
        divisor : Int
            Distance between each strike in dollars. The default is 25 for SPX
            and 10 otherwise.
        r : Float
            Interest Rate. The default is 0.005.
        q : Float
            Dividend Yield. The default is 0.
        epsilon : Float
            Degree of precision to return implied vol. The default
            is 0.001.
        method : Str
            Implied Vol method; 'nr', 'bisection' or 'naive'. The
            default is 'nr'.

        Returns
        -------
        DataFrame
            DataFrame of Option data.

        """

        # Dictionary of parameter defaults
        self.df_dict = copy.deepcopy(vol_params_dict)

        # Store initial inputs
        inputs = {}
        for key, value in kwargs.items():
            inputs[key] = value

        # Initialise system parameters
        params = Utils.init_params(inputs)
        tables = {}

        params, tables = Data.create_option_data(params=params, tables=tables)

        self.params = params
        self.tables = tables


    def visualize(self, **kwargs):
        """
        Visualize the implied volatility as 2D linegraph, 3D scatter
        or 3D surface

        Parameters
        ----------
        graphtype : Str
            Whether to display 'line', 'scatter' or 'surface'. The
            default is 'line'.
        surfacetype : Str
            The type of 3D surface to display from 'trisurf', 'mesh',
            spline', 'interactive_mesh' and 'interactive_spline'. The
            default is 'mesh'.
        smoothing : Bool
            Whether to apply polynomial smoothing. The default is False.
        scatter : Bool
            Whether to plot scatter points on 3D mesh grid. The default
            is False.
        voltype : Str
            Whether to use 'bid', 'mid', 'ask' or 'last' price. The
            default is 'last'.
        order : Int
            Polynomial order used in numpy polyfit function. The
            default is 3.
        spacegrain : Int
            Number of points in each axis linspace argument for 3D
            graphs. The default is 100.
        azim : Float
            L-R view angle for 3D graphs. The default is -50.
        elev : Float
            Elevation view angle for 3D graphs. The default is 20.
        fig_size : Tuple
            3D graph size
        rbffunc : Str
            Radial basis function used in interpolation chosen from
            'multiquadric', 'inverse', 'gaussian', 'linear', 'cubic',
            'quintic', 'thin_plate'. The default is 'thin_plate'
        colorscale : Str
            Colors used in plotly interactive graph. The default is
            'BlueRed'
        opacity : Float
            opacity of 3D interactive graph
        surf : Bool
            Plot surface in interactive graph
        notebook : Bool
            Whether interactive graph is run in Jupyter notebook or
            IDE. The default is False.
        save_image : Bool
            Whether to save a copy of the image as a png file. The default is
            False
        image_folder : Str
            Location to save any images. The default is 'images'
        image_dpi : Int
            Resolution to save images. The default is 50.

        Returns
        -------
        Displays the output of the chosen graphing method.

        """

        # Update params with the specified parameters
        for key, value in kwargs.items():

            # Replace the default parameter with that provided
            self.params[key] = value

        # Run method selected by graphtype
        if self.params['graphtype'] == 'line':
            self.params, self.tables = Graph.line_graph(
                params=self.params, tables=self.tables)

        elif self.params['graphtype'] == 'scatter':
            self.params, self.tables = Graph.scatter_3d(
                params=self.params, tables=self.tables)

        elif self.params['graphtype'] == 'surface':
            self.params, self.tables = Graph.surface_3d(
                params=self.params, tables=self.tables)

        else:
            print ("Please select a graphtype from 'line', "\
                   "'scatter' and 'surface'")

    def linegraph(self, **kwargs):
        """
        Displays a linegraph of each option maturity plotted by strike
        and implied vol

        Parameters
        ----------
        voltype : Str
            Whether to use 'bid', 'mid', 'ask' or 'last' price. The
            default is 'last'.
        save_image : Bool
            Whether to save a copy of the image as a png file. The default is
            False
        image_folder : Str
            Location to save any images. The default is 'images'
        image_dpi : Int
            Resolution to save images. The default is 50.

        Returns
        -------
        Linegraph.

        """

        # Update params with the specified parameters
        for key, value in kwargs.items():

            # Replace the default parameter with that provided
            self.params[key] = value

        self.params, self.tables = Graph.line_graph(
            params=self.params, tables=self.tables)


    def scatter(self, **kwargs):
        """
        Displays a 3D scatter plot of each option implied vol against
        strike and maturity

        Parameters
        ----------
        voltype : Str
            Whether to use 'bid', 'mid', 'ask' or 'last' price. The
            default is 'last'.
        azim : Float
            L-R view angle for 3D graphs. The default is -50.
        elev : Float
            Elevation view angle for 3D graphs. The default is 20.
        fig_size : Tuple
            3D graph size
        save_image : Bool
            Whether to save a copy of the image as a png file. The default is
            False
        image_folder : Str
            Location to save any images. The default is 'images'
        image_dpi : Int
            Resolution to save images. The default is 50.

        Returns
        -------
        3D Scatter plot.

        """

        # Update params with the specified parameters
        for key, value in kwargs.items():

            # Replace the default parameter with that provided
            self.params[key] = value

        self.params, self.tables = Graph.scatter_3d(
            params=self.params, tables=self.tables)


    def surface(self, **kwargs):
        """
        Displays a 3D surface plot of the implied vol surface against
        strike and maturity

        Parameters
        ----------
        surfacetype : Str
            The type of 3D surface to display from 'trisurf', 'mesh',
            'spline', 'interactive_mesh' and 'interactive_spline'.
            The default is 'mesh'.
        smoothing : Bool
            Whether to apply polynomial smoothing. The default is False.
        scatter : Bool
            Whether to plot scatter points on 3D mesh grid. The default
            is False.
        voltype : Str
            Whether to use 'bid', 'mid', 'ask' or 'last' price. The
            default is 'last'.
        order : Int
            Polynomial order used in numpy polyfit function. The
            default is 3.
        spacegrain : Int
            Number of points in each axis linspace argument for 3D
            graphs. The default
            is 100.
        azim : Float
            L-R view angle for 3D graphs. The default is -50.
        elev : Float
            Elevation view angle for 3D graphs. The default is 20.
        fig_size : Tuple
            3D graph size
        rbffunc : Str
            Radial basis function used in interpolation chosen from
            'multiquadric', 'inverse', 'gaussian', 'linear', 'cubic',
            'quintic', 'thin_plate'. The default is 'thin_plate'
        colorscale : Str
            Colors used in plotly interactive graph. The default is
            'BlueRed'
        opacity : Float
            opacity of 3D interactive graph
        surf : Bool
            Plot surface in interactive graph
        notebook : Bool
            Whether interactive graph is run in Jupyter notebook or IDE.
            The default is False.
        save_image : Bool
            Whether to save a copy of the image as a png file. The default is
            False
        image_folder : Str
            Location to save any images. The default is 'images'
        image_dpi : Int
            Resolution to save images. The default is 50.

        Returns
        -------
        3D surface plot.

        """

        # Update params with the specified parameters
        for key, value in kwargs.items():

            # Replace the default parameter with that provided
            self.params[key] = value

        self.params, self.tables = Graph.surface_3d(
            params=self.params, tables=self.tables)



"""
Market data import and transformation functions

"""
import calendar
from collections import Counter
import copy
from datetime import date
import time
from urllib.request import FancyURLopener
import warnings
import datetime as dt
from bs4 import BeautifulSoup
from lxml import html
import numpy as np
import pandas as pd
import pytz
from volvisualizer.utils import ImpliedVol
# pylint: disable=invalid-name

# Class used to open urls for financial data
class UrlOpener(FancyURLopener):
    """
    Extract data from Yahoo Finance URL

    """
    version = 'w3m/0.5.3+git20180125'


class Data():
    """
    Market data import and transformation functions

    """
    @classmethod
    def create_option_data(cls, params, tables):
        """
        Extract the URL for each of the listed option on Yahoo Finance
        for the given ticker. Extract option data from each URL.

        Filter / transform the data and calculate implied volatilities for
        specified put and call strikes.

        Parameters
        ----------
        start_date : Str
            Date from when to include prices (some of the options
            won't have traded for days / weeks and therefore will
            have stale prices).
        ticker : Str
            The ticker identifier used by Yahoo for the chosen stock. The
            default is '^SPX'.
        ticker_label : Str
            The ticker label used in charts.
        wait : Int
            Number of seconds to wait between each url query
        lastmins : Int, Optional
            Restrict to trades within number of minutes since last
            trade time recorded
        mindays : Int, Optional
            Restrict to options greater than certain option expiry
        minopts : Int, Optional
            Restrict to minimum number of options to include that
            option expiry
        volume : Int, Optional
            Restrict to minimum Volume
        openint : Int, Optional
            Restrict to minimum Open Interest
        monthlies : Bool
            Restrict expiries to only 3rd Friday of the month. Default
            is False.
        spot : Float
            Underlying reference level.
        put_strikes : List
            Range of put strikes to calculate implied volatility for.
        call_strikes : List
            Range of call strikes to calculate implied volatility for.
        strike_limits : Tuple
            min and max strikes to use expressed as a decimal
            percentage. The default is (0.5, 2.0).
        divisor : Int
            Distance between each strike in dollars. The default is 25 for SPX
            and 10 otherwise.
        r : Float
            Interest Rate. The default is 0.005.
        q : Float
            Dividend Yield. The default is 0.
        epsilon : Float
            Degree of precision to return implied vol. The default
            is 0.001.
        method : Str
            Implied Vol method; 'nr', 'bisection' or 'naive'. The
            default is 'nr'.

        Returns
        -------
        DataFrame
            DataFrame of Option data.

        """
        # Extract URLs and option data
        params, tables = cls.extractoptions(params=params, tables=tables)
        print("Options data extracted")

        # Filter / transform data
        params, tables = cls.transform(params=params, tables=tables)
        print("Data transformed")

        # Calculate implied volatilities and combine
        params, tables = cls.combine(params=params, tables=tables)
        print("Data combined")

        return params, tables


    @classmethod
    def extractoptions(cls, params, tables):
        """
        Extract the URL for each of the listed option on Yahoo Finance
        for the given ticker. Extract option data from each URL.


        Parameters
        ----------
        ticker : Str
            The ticker identifier used by Yahoo for the chosen stock. The
            default is '^SPX'.
        wait : Int
            Number of seconds to wait between each url query

        Returns
        -------
        DataFrame
            All option data from each of the supplied urls.

        """

        # Extract dictionary of option dates and urls
        params = cls._extracturls(params=params)
        print("URL's extracted")

        params['raw_web_data'] = cls._extract_web_data(params=params)

        params = cls._read_web_data(params=params)

        # Create an empty DataFrame
        tables['full_data'] = pd.DataFrame()

        # Make a list of all the dates of the DataFrames just stored
        # in the default dictionary
        params['date_list'] = list(params['option_dict'].keys())

        params, tables = cls._process_options(params=params, tables=tables)

        return params, tables


    @staticmethod
    def _extracturls(params):
        """
        Extract the URL for each of the listed option on Yahoo Finance
        for the given ticker.

        Parameters
        ----------
        ticker : Str
            Yahoo ticker (Reuters RIC) for the stock.

        Returns
        -------
        Dict
            Dictionary of dates and URLs.

        """

        # Define the stock root webpage
        url = 'https://finance.yahoo.com/quote/'+params['ticker']\
            +'/options?p='+params['ticker']

        # Create a UrlOpener object to extract data from the url
        urlopener = UrlOpener()
        response = urlopener.open(url)

        # Collect the text from this object
        params['html_doc'] = response.read()

        # Use Beautiful Soup to parse this
        soup = BeautifulSoup(params['html_doc'], features="lxml")

        # Create a list of all the option dates
        option_dates = [a.get_text() for a in soup.find_all('option')]

        # Convert this list from string to datetimes
        dates_list = [dt.datetime.strptime(date, "%B %d, %Y").date() for date
                      in option_dates]

        # Convert back to strings in the required format
        str_dates = [date_obj.strftime('%Y-%m-%d') for date_obj in dates_list]

        # Create a list of all the unix dates used in the url for each
        # of these dates
        option_pages = [a.attrs['value'] for a in soup.find_all('option')]

        # Combine the dates and unixdates in a dictionary
        optodict = dict(zip(str_dates, option_pages))

        # Create an empty dictionary
        params['url_dict'] = {}

        # For each date and unixdate in the first dictionary
        for date_val, page in optodict.items():

            # Create an entry with the date as key and the url plus
            # unix date as value
            params['url_dict'][date_val] = str(
                'https://finance.yahoo.com/quote/'
                +params['ticker']+'/options?date='+page)

        return params



    @staticmethod
    def _extract_web_data(params):

        raw_web_data = {}

        # each url needs to have an option expiry date associated with
        # it in the url dict
        for input_date, url in params['url_dict'].items():

            # UrlOpener function downloads the data
            urlopener = UrlOpener()
            weburl = urlopener.open(url)
            raw_web_data[input_date] = weburl.read()

            # wait between each query so as not to overload server
            time.sleep(params['wait'])

        return raw_web_data


    @staticmethod
    def _read_web_data(params):

        # Create an empty dictionary
        params['option_dict'] = {}
        params['url_except_dict'] = {}

        for input_date, url in params['url_dict'].items():
            # if data exists
            try:
                # read html data into a DataFrame
                option_frame = pd.read_html(params['raw_web_data'][input_date])

                # Add this DataFrame to the default dictionary, named
                # with the expiry date it refers to
                params['option_dict'][input_date] = option_frame

            # otherwise collect dictionary of exceptions
            except ValueError:
                params['url_except_dict'][input_date] = url

        return params


    @staticmethod
    def _process_options(params, tables):

        # Create list to store exceptions
        params['opt_except_list'] = []

        # For each of these dates
        for input_date in params['date_list']:

            try:
                # The first entry is 'calls'
                calls = params['option_dict'][input_date][0]

                # Create a column designating these as calls
                calls['Option Type'] = 'call'

                try:
                    # The second entry is 'puts'
                    puts = params['option_dict'][input_date][1]

                    # Create a column designating these as puts
                    puts['Option Type'] = 'put'

                    # Concatenate these two DataFrames
                    options = pd.concat([calls, puts])

                    # Add an 'Expiry' column with the expiry date
                    options['Expiry'] = pd.to_datetime(input_date).date()

                    # Add this DataFrame to 'full_data'
                    tables['full_data'] = pd.concat(
                        [tables['full_data'], options])

                except IndexError:

                    # Add an 'Expiry' column with the expiry date
                    calls['Expiry'] = pd.to_datetime(input_date).date()

                    # Add this DataFrame to 'full_data'
                    tables['full_data'] = pd.concat(
                        [tables['full_data'], calls])

            except IndexError:

                try:
                    # The second entry is 'puts'
                    puts = params['option_dict'][input_date][1]

                    # Create a column designating these as puts
                    puts['Option Type'] = 'put'

                    # Add an 'Expiry' column with the expiry date
                    puts['Expiry'] = pd.to_datetime(input_date).date()

                    # Add this DataFrame to 'full_data'
                    tables['full_data'] = pd.concat(
                        [tables['full_data'], puts])

                except IndexError:
                    params['opt_except_list'].append(input_date)

        return params, tables


    @classmethod
    def transform(cls, params, tables):
        """
        Perform some filtering / transforming of the option data

        Parameters
        ----------
        start_date : Str
            Date from when to include prices (some of the options
            won't have traded for days / weeks and therefore will
            have stale prices).
        lastmins : Int, Optional
            Restrict to trades within number of minutes since last
            trade time recorded
        mindays : Int, Optional
            Restrict to options greater than certain option expiry
        minopts : Int, Optional
            Restrict to minimum number of options to include that
            option expiry
        volume : Int, Optional
            Restrict to minimum Volume
        openint : Int, Optional
            Restrict to minimum Open Interest
        monthlies : Bool
            Restrict expiries to only 3rd Friday of the month. Default
            is False.

        Returns
        -------
        DataFrame
            Creates a new DataFrame as a modification of 'full_data'.

        """

        # Make a copy of 'full_data'
        tables['data'] = copy.deepcopy(tables['full_data'])

        # Set timezone
        est = pytz.timezone('US/Eastern')

        # Convert 'Last Trade Date' to a DateTime variable
        tables['data']['Last Trade Date Raw'] = (
            tables['data']['Last Trade Date'])

        # Format date based on Eastern Daylight or Standard Time
        try:
            tables['data']['Last Trade Date'] = pd.to_datetime(
                tables['data']['Last Trade Date'],
                format='%Y-%m-%d %I:%M%p EDT')

        except KeyError:
            tables['data']['Last Trade Date'] = pd.to_datetime(
                tables['data']['Last Trade Date'],
                format='%Y-%m-%d %I:%M%p EST')

        tables['data']['Last Trade Date'] = (
            tables['data']['Last Trade Date'].apply(
                lambda x: x.replace(tzinfo=est)))

        # Create columns of expiry date as datetime object and str
        tables['data']['Expiry_datetime'] = pd.to_datetime(
            tables['data']['Expiry'], format='%Y-%m-%d')
        tables['data']['Expiry_str'] = (
            tables['data']['Expiry_datetime'].dt.strftime('%Y-%m-%d'))

        # Filter data from start date
        tables['data'] = (
            tables['data'][tables['data']['Last Trade Date']>=str(
                pd.to_datetime(params['start_date']))])

        tables = cls._trade_columns(tables=tables)

        # Create Time To Maturity (in years) column
        tables['data']['TTM'] = (
            pd.to_datetime(tables['data']['Expiry'])
            - pd.to_datetime(date.today())) / (pd.Timedelta(days=1) * 365)

        # Create Days to Maturity column
        tables['data']['Days'] = np.round(tables['data']['TTM']*365, 0)

        params, tables = cls._filters(params=params, tables=tables)

        return params, tables


    @staticmethod
    def _trade_columns(tables):

        # Create a column of the Trade Day
        tables['data']['Last Trade Day'] = (
            tables['data']['Last Trade Date'].dt.date)

        # Create a column of the Trade Time of Day
        tables['data']['Last Trade Time'] = (
            tables['data']['Last Trade Date'].dt.time)

        # Create a column of the Trade Hour of Day
        tables['data']['Last Trade Hour'] = (
            tables['data']['Last Trade Date'].dt.hour)

        # Create a column of the Trade Date represented in unixtime
        tables['data']['Unixtime'] = (
            tables['data']['Last Trade Date'].view(np.int64) // 10**9)

        # Clean Volume column
        tables['data']['Volume'] = (
            tables['data']['Volume'].replace('-',0).astype(int))

        # Clean Open Interest column
        tables['data']['Open Interest'] = (
            tables['data']['Open Interest'].replace('-',0).astype(int))

        # Clean Ask column
        tables['data']['Ask'] = (
            tables['data']['Ask'].replace('-',0).astype(float))

        # Clean Bid column
        tables['data']['Bid'] = (
            tables['data']['Bid'].replace('-',0).astype(float))

        # Create Mid column
        tables['data']['Mid'] = (
            tables['data']['Ask'] + tables['data']['Bid']) / 2

        return tables


    @classmethod
    def _filters(cls, params, tables):

        # If a minutes parameter is supplied, filter for most recent
        # minutes
        if params['lastmins'] is not None:
            tables['data'] = (tables['data'][tables['data']['Unixtime']  >= (
                max(tables['data']['Unixtime']) - params['lastmins'] * 60)])

        # If a mindays parameter is supplied, filter for option expiry
        # greater than parameter
        if params['mindays'] is not None:
            tables['data'] = (
                tables['data'][tables['data']['Days']  >= params['mindays']])

        # If a minopts parameter is supplied, filter for volume greater
        # than parameter
        if params['minopts'] is not None:

            # Create a dictionary of the number of options for each
            # maturity
            mat_dict = dict(Counter(tables['data']['Days']))
            for ttm, count in mat_dict.items():

                # if there are less than minopts options for a given
                # maturity
                if count < params['minopts']:

                    # remove that maturity from the DataFrame
                    tables['data'] = (
                        tables['data'][tables['data']['Days'] != ttm])

        # If a volume parameter is supplied, filter for volume greater
        # than parameter
        if params['volume'] is not None:
            tables['data'] = (
                tables['data'][tables['data']['Volume']  >= params['volume']])

        # If an openint parameter is supplied, filter for Open Interest
        # greater than parameter
        if params['openint'] is not None:
            tables['data'] = (
                tables['data'][tables['data']['Open Interest']
                               >= params['openint']])

        params, tables = cls._monthlies(params=params, tables=tables)

        return params, tables


    @staticmethod
    def _monthlies(params, tables):

        # If the monthlies flag is set
        if params['monthlies'] is True:

            # Create an empty list
            date_list = []

            # For each date in the url_dict
            for key in params['url_dict'].keys():

                # Format that string as a datetime object
                key_date = dt.datetime.strptime(key, "%Y-%m-%d")

                # Store the year and month as a tuple in date_list
                date_list.append((key_date.year, key_date.month))

            # Create a sorted list from the unique dates in date_list
            sorted_dates = sorted(list(set(date_list)))

            # Create an empty list
            days_list = []

            # Create a calendar object
            c = calendar.Calendar(firstweekday=calendar.SATURDAY)

            # For each tuple of year, month in sorted_dates
            for tup in sorted_dates:

                # Create a list of lists of days in that month
                monthcal = c.monthdatescalendar(tup[0], tup[1])

                # Extract the date corresponding to the 3rd Friday
                expiry = monthcal[2][-1]

                # Calculate the number of days until that expiry
                ttm = (expiry - dt.date.today()).days

                # Append this to the days_list
                days_list.append(ttm)

            # For each unique number of days to expiry
            for days_to_expiry in set(tables['data']['Days']):

                # if the expiry is not in the list of monthly expiries
                if days_to_expiry not in days_list:

                    # Remove that expiry from the DataFrame
                    tables['data'] = (
                        tables['data'][tables['data']['Days']
                                       != days_to_expiry])

        return params, tables


    @classmethod
    def combine(cls, params, tables):
        """
        Calculate implied volatilities for specified put and call
        strikes and combine.

        Parameters
        ----------
        ticker_label : Str
            Stock Ticker.
        spot : Float
            Underlying reference level.
        put_strikes : List
            Range of put strikes to calculate implied volatility for.
        call_strikes : List
            Range of call strikes to calculate implied volatility for.
        strike_limit : Tuple
            min and max strikes to use expressed as a decimal
            percentage. The default is (0.5, 2.0).
        divisor : Int
            Distance between each strike in dollars. The default is 25 for SPX
            and 10 otherwise.
        r : Float
            Interest Rate. The default is 0.005.
        q : Float
            Dividend Yield. The default is 0.
        epsilon : Float
            Degree of precision to return implied vol. The default
            is 0.001.
        method : Str
            Implied Vol method; 'nr', 'bisection' or 'naive'. The
            default is 'nr'.

        Returns
        -------
        DataFrame
            DataFrame of Option prices.

        """

        # create copy of filtered data
        input_data = copy.deepcopy(tables['data'])

        # Calculate strikes if strikes and spot price are not supplied.
        params = cls._create_strike_range(params=params, tables=tables)

        # Assign ticker label to the object
        if params['ticker_label'] is None:
            params['ticker_label'] = params['ticker'].lstrip('^')

        # Create empty list and dictionary for storing options
        params['opt_list'] = []
        opt_dict = {}

        # For each put strike price
        for strike in params['put_strikes']:

            # Assign an option name of ticker plus strike
            opt_name = params['ticker_label']+'_'+str(strike)

            # store the implied vol results for that strike in the
            # option dictionary
            opt_dict[opt_name] = cls._imp_vol_apply(
                params=params, input_data=input_data, strike=strike,
                option='put')

            # store the implied vol results for that strike in the
            # option list
            params['opt_list'].append(opt_dict[opt_name])

            print('Put option: ', opt_name)

        # For each put strike price
        for strike in params['call_strikes']:

            # Assign an option name of ticker plus strike
            opt_name = params['ticker_label']+'_'+str(strike)

            # store the implied vol results DataFrame for that strike
            # in the option dictionary
            opt_dict[opt_name] = cls._imp_vol_apply(
                params=params, input_data=input_data, strike=strike,
                option='call')

            # store the implied vol results DataFrame for that strike
            # in the option list
            params['opt_list'].append(opt_dict[opt_name])

            print('Call option: ', opt_name)

        # Concatenate all the option results into a single DataFrame and drop
        # any null values
        tables['imp_vol_data'] = pd.concat(params['opt_list']).dropna()

        return params, tables


    @classmethod
    def _create_strike_range(cls, params, tables):

        # Extract the spot level from the html data
        if params['spot'] is None:
            tree = html.fromstring(params['html_doc'])
            priceparse = tree.xpath(
                '//span[@class="Trsdu(0.3s) Fw(b) Fz(36px) Mb(-4px) D(ib)"]/text()')
            params['spot'] = float(
                [str(p) for p in priceparse][0].replace(',',''))

        # Calculate initial spot, min and max strikes to use in divisor
        # calculation
        params['init_roundspot'], params['init_put_min'], \
            params['init_call_max'] = cls._strike_filters(
                params=params, divisor=10)

        # If a divisor value is not provided
        if params['divisor'] is None:
            params = cls._create_divisor(params=params, tables=tables)

        # Calculate final spot, min and max strikes using chosen divisor
        params['roundspot'], params['put_min'], \
            params['call_max'] = cls._strike_filters(
                params=params, divisor=params['divisor'])

        # Calculate put options (default is 1/2 spot level)
        if params['put_strikes'] is None:
            params['put_strikes'] = list(
                np.linspace(
                    params['put_min'],
                    params['roundspot'],
                    int((params['roundspot'] - params['put_min'])
                        / params['divisor']) + 1))

        # Calculate call options (default is twice the spot level)
        if params['call_strikes'] is None:
            params['call_strikes'] = list(
                np.linspace(
                    params['roundspot'],
                    params['call_max'],
                    int((params['call_max'] - params['roundspot'])
                        / params['divisor']) + 1))

        return params


    @staticmethod
    def _strike_filters(params, divisor):

        # Calculate the point to switch from put to call options
        roundspot = (
            round(params['spot'] / divisor) * divisor)

        put_min = (
            round(params['spot'] * params['strike_limits'][0] / divisor)
            * divisor)

        call_max = (
            round(params['spot'] * params['strike_limits'][1] / divisor)
            * divisor)

        return roundspot, put_min, call_max


    @staticmethod
    def _create_divisor(params, tables):

        # Take the set of all the option strikes in the data
        strikes = set(tables['data']['Strike'])

        # Find the number of options with a remainder of zero for each of the
        # listed potential divisors
        avail_strikes = {}
        for div in [0.5, 1, 1.25, 2.5, 5, 10, 25, 50, 100]:
            avail_strikes[div] = len(
                {x for x in strikes if (
                    x%(div) == 0
                    and params['init_put_min'] < x < params['init_call_max'])})

        # Find the maximum divisor of those with the highest number of options
        params['divisor'] = max(
            [divisor for max_strike_count in [max(avail_strikes.values())]
             for divisor, strike_count in avail_strikes.items()
             if strike_count == max_strike_count])

        return params


    @classmethod
    def _imp_vol_apply(cls, params, input_data, strike, option):
        """
        Apply _implied_vol_by_row method to each row of a DataFrame.

        Parameters
        ----------
        input_data : DataFrame
            DataFrame of Option prices.
        S : Float
            Stock Price.
        K : Float
            Strike Price.
        r : Float
            Interest Rate. The default is 0.005.
        q : Float
            Dividend Yield. The default is 0.
        epsilon : Float
            Degree of precision to return implied vol. The default is
            0.001.
        option : Str
            Option type; 'put' or 'call'.
        method : Str
            Implied Vol method; 'nr', 'bisection' or 'naive'. The
            default is 'nr'.

        Returns
        -------
        input_data : DataFrame
            DataFrame of Option prices.

        """

        # Filter data by strike and option type
        input_data = (input_data[(input_data['Strike'] == strike)
                & (input_data['Option Type'] == option)])

        # Apply implied vol method to each row
        input_data = input_data.apply(
            lambda x: cls._imp_vol_by_row(x, params, strike, option), axis=1)

        return input_data


    @staticmethod
    def _imp_vol_by_row(row, params, strike, option):
        """
        Calculate implied vol for one row of a DataFrame.

        Parameters
        ----------
        row : Array
            Each row in the DataFrame.
        S : Float
            Stock Price.
        K : Float
            Strike Price.
        r : Float
            Interest Rate.
        q : Float
            Dividend Yield.
        epsilon : Float
            Degree of precision to return implied vol.
        option : Str
            Option type; 'put' or 'call'.
        method : Str
            Implied Vol method; 'newtonraphson', 'bisection' or
            'iv_naive'.

        Returns
        -------
        row : Array
            Each row in the DataFrame.

        """

        # Suppress runtime warnings caused by bad vol data
        warnings.filterwarnings("ignore", category=RuntimeWarning)


        # Select the chosen implied vol method from the dictionary
        func_name = params['method_dict'][params['method']]

        # for each of the prices: bid, mid, ask, last
        for input_row, output_row in params['row_dict'].items():

            opt_params = {
                'S':params['spot'],
                'K':strike,
                'T':row['TTM'],
                'r':params['r'],
                'q':params['q'],
                'cm':row[input_row],
                'epsilon':params['epsilon'],
                'option':option
                }
            try:
                # populate the column using the chosen implied
                # vol method (using getattr() to select
                # dynamically)
                # check if n/a value is returned and print error
                # message if so
                output = getattr(
                    ImpliedVol, func_name)(opt_params=opt_params)

                output = float(output)
                row[output_row] = output

            except KeyError:
                print("Error with option: Strike="+str(strike)+
                          " TTM="+str(round(row['TTM'], 3))+
                          " vol="+str(row[input_row])+
                          " option="+option)

        # Return warnings to default setting
        warnings.filterwarnings("default", category=RuntimeWarning)

        return row



"""
Key parameters for visualizer

"""

# Dictionary of default parameters
vol_params_dict = {
    'ticker':'^SPX',
    'ticker_label':None,
    'start_date':None,
    'wait':5,
    'minopts':None,
    'mindays':None,
    'lastmins':None,
    'volume':None,
    'openint':None,
    'graphtype':'line',
    'surfacetype':'mesh',
    'smoothing':False,
    'scatter':True,
    'voltype':'last',
    'smoothopt':6,
    'notebook':False,
    'r':0.005,
    'q':0,
    'epsilon':0.001,
    'method':'nr',
    'order':3,
    'spacegrain':100,
    'azim':-50,
    'elev':20,
    'fig_size':(15, 12),
    'rbffunc':'thin_plate',
    'colorscale':'BlueRed',
    'monthlies':True,
    'divisor':None,
    'divisor_SPX':25,
    'spot':None,
    'strike_limits':(0.5, 2.0),
    'put_strikes':None,
    'call_strikes':None,
    'opacity':1,
    'surf':True,
    'save_image':False,
    'image_folder':'images',
    'image_filename':'impvol',
    'image_dpi':50,

    # Dictionary of implied vol fields used in graph methods
    'vols_dict':{
        'bid':'Imp Vol - Bid',
        'mid':'Imp Vol - Mid',
        'ask':'Imp Vol - Ask',
        'last':'Imp Vol - Last'
        },

    # Dictionary of price fields used for filtering zero prices in
    # graph methods
    'prices_dict':{
        'bid':'Bid',
        'mid':'Mid',
        'ask':'Ask',
        'last':'Last Price'
        },

    # Dictionary of implied vol fields used in implied vol calculation
    'row_dict':{
        'Bid':'Imp Vol - Bid',
        'Mid':'Imp Vol - Mid',
        'Ask':'Imp Vol - Ask',
        'Last Price':'Imp Vol - Last'
        },

    # Dictionary of interpolation methods used in implied vol calculation
    'method_dict':{
        'nr':'implied_vol_newton_raphson',
        'bisection':'implied_vol_bisection',
        'naive':'implied_vol_naive'
        },

    # Parameters to overwrite mpl_style defaults
    'mpl_line_params':{
        'axes.edgecolor':'black',
        'axes.titlepad':20,
        'axes.xmargin':0.05,
        'axes.ymargin':0.05,
        'axes.linewidth':2,
        'axes.facecolor':(0.8, 0.8, 0.9, 0.5),
        'xtick.major.pad':10,
        'ytick.major.pad':10,
        'lines.linewidth':3.0,
        'grid.color':'black',
        'grid.linestyle':':'
        },

    'mpl_3D_params':{
        'axes.facecolor':'w',
        'axes.labelcolor':'k',
        'axes.edgecolor':'w',
        'lines.linewidth':0.5,
        'xtick.labelbottom':True,
        'ytick.labelleft':True
        },

    }



"""
Utility functions for refreshing parameters and getting implied volatility
from prices

"""
import copy
import datetime as dt
import numpy as np
import scipy.stats as si
from pandas.tseries.offsets import BDay
from volvisualizer.volatility_params import vol_params_dict
# pylint: disable=invalid-name

class Utils():
    """
    Utility functions for refreshing parameters

    """
    @classmethod
    def init_params(cls, inputs):
        """
        Initialise parameter dictionary
        Parameters
        ----------
        inputs : Dict
            Dictionary of parameters supplied to the function.
        Returns
        -------
        params : Dict
            Dictionary of parameters.
        """
        # Copy the default parameters
        params = copy.deepcopy(vol_params_dict)

        # For all the supplied arguments
        for key, value in inputs.items():

            # Replace the default parameter with that provided
            params[key] = value

        params = cls.set_start_date(params=params)

        return params


    @staticmethod
    def set_start_date(params):
        """
        Set start date to previous working day if not provided

        Parameters
        ----------
        params : Dict
            Dictionary of key parameters.

        Returns
        -------
        params : Dict
            Dictionary of key parameters.

        """
        if params['start_date'] is None:
            start_date_as_dt = (dt.datetime.today() - BDay(1)).date()
            params['start_date'] = str(start_date_as_dt)

        return params


class ImpliedVol():
    """
    Implied Volatility Extraction methods

    """
    @classmethod
    def implied_vol_newton_raphson(cls, opt_params):
        """
        Finds implied volatility using Newton-Raphson method - needs
        knowledge of partial derivative of option pricing formula
        with respect to volatility (vega)

        Parameters
        ----------
        S : Float
            Stock Price. The default is 100.
        K : Float
            Strike Price. The default is 100.
        T : Float
            Time to Maturity.  The default is 0.25 (3 Months).
        r : Float
            Interest Rate. The default is 0.005 (50bps)
        q : Float
            Dividend Yield.  The default is 0.
        cm : Float
            # Option price used to solve for vol. The default is 5.
        epsilon : Float
            Degree of precision. The default is 0.0001
        option : Str
            Type of option. 'put' or 'call'. The default is 'call'.
        default : Bool
            Whether the function is being called directly (in which
            case values that are not supplied are set to default
            values) or called from another function where they have
            already been updated.

        Returns
        -------
        result : Float
            Implied Volatility.

        """

        # Manaster and Koehler seed value
        opt_params['vi'] = np.sqrt(
            abs(np.log(opt_params['S'] / opt_params['K'])
                + opt_params['r'] * opt_params['T']) * (2 / opt_params['T']))

        opt_params['ci'] = cls.black_scholes_merton(
            opt_params=opt_params, sigma=opt_params['vi'])

        opt_params['vegai'] = cls.black_scholes_merton_vega(
            opt_params=opt_params, sigma=opt_params['vi'])

        opt_params['mindiff'] = abs(opt_params['cm'] - opt_params['ci'])

        while (abs(opt_params['cm'] - opt_params['ci'])
               >= opt_params['epsilon']
               and abs(opt_params['cm'] - opt_params['ci'])
               <= opt_params['mindiff']):

            opt_params['vi'] = (
                opt_params['vi']
                - (opt_params['ci'] - opt_params['cm']) / opt_params['vegai'])

            opt_params['ci'] = cls.black_scholes_merton(
                opt_params=opt_params, sigma=opt_params['vi'])

            opt_params['vegai'] = cls.black_scholes_merton_vega(
                opt_params=opt_params, sigma=opt_params['vi'])

            opt_params['mindiff'] = abs(opt_params['cm'] - opt_params['ci'])

        if abs(opt_params['cm'] - opt_params['ci']) < opt_params['epsilon']:
            result = opt_params['vi']
        else:
            result = 'NA'

        return result


    @classmethod
    def implied_vol_bisection(cls, opt_params):
        """
        Finds implied volatility using bisection method.

        Parameters
        ----------
        S : Float
            Stock Price. The default is 100.
        K : Float
            Strike Price. The default is 100.
        T : Float
            Time to Maturity.  The default is 0.25 (3 Months).
        r : Float
            Interest Rate. The default is 0.005 (50bps)
        q : Float
            Dividend Yield.  The default is 0.
        cm : Float
            # Option price used to solve for vol. The default is 5.
        epsilon : Float
            Degree of precision. The default is 0.0001
        option : Str
            Type of option. 'put' or 'call'. The default is 'call'.
        default : Bool
            Whether the function is being called directly (in which
            case values that are not supplied are set to default
            values) or called from another function where they have
            already been updated.

        Returns
        -------
        result : Float
            Implied Volatility.

        """

        opt_params['vLow'] = 0.005
        opt_params['vHigh'] = 4
        opt_params['cLow'] = cls.black_scholes_merton(
            opt_params=opt_params, sigma=opt_params['vLow'])

        opt_params['cHigh'] = cls.black_scholes_merton(
            opt_params=opt_params, sigma=opt_params['vHigh'])

        counter = 0

        opt_params['vi'] = (
            opt_params['vLow']
            + (opt_params['cm'] - opt_params['cLow'])
            * (opt_params['vHigh'] - opt_params['vLow'])
            / (opt_params['cHigh'] - opt_params['cLow']))

        while abs(opt_params['cm'] - cls.black_scholes_merton(
                opt_params=opt_params,
                sigma=opt_params['vi'])) > opt_params['epsilon']:

            counter = counter + 1
            if counter == 100:
                result = 'NA'

            if cls.black_scholes_merton(
                    opt_params=opt_params,
                    sigma=opt_params['vi']) < opt_params['cm']:
                opt_params['vLow'] = opt_params['vi']

            else:
                opt_params['vHigh'] = opt_params['vi']

            opt_params['cLow'] = cls.black_scholes_merton(
                opt_params=opt_params, sigma=opt_params['vLow'])

            opt_params['cHigh'] = cls.black_scholes_merton(
                opt_params=opt_params, sigma=opt_params['vHigh'])

            opt_params['vi'] = (
                opt_params['vLow']
                + (opt_params['cm'] - opt_params['cLow'])
                * (opt_params['vHigh'] - opt_params['vLow'])
                / (opt_params['cHigh'] - opt_params['cLow']))

        result = opt_params['vi']

        return result


    @classmethod
    def implied_vol_naive(cls, opt_params):
        """
        Finds implied volatility using simple naive iteration,
        increasing precision each time the difference changes sign.

        Parameters
        ----------
        S : Float
            Stock Price. The default is 100.
        K : Float
            Strike Price. The default is 100.
        T : Float
            Time to Maturity.  The default is 0.25 (3 Months).
        r : Float
            Interest Rate. The default is 0.005 (50bps)
        q : Float
            Dividend Yield.  The default is 0.
        cm : Float
            # Option price used to solve for vol. The default is 5.
        epsilon : Float
            Degree of precision. The default is 0.0001
        option : Str
            Type of option. 'put' or 'call'. The default is 'call'.

        Returns
        -------
        result : Float
            Implied Volatility.

        """

        # Seed vol
        opt_params['vi'] = 0.2

        # Calculate starting option price using this vol
        opt_params['ci'] = cls.black_scholes_merton(
            opt_params=opt_params, sigma=opt_params['vi'])

        # Initial price difference
        opt_params['price_diff'] = opt_params['cm'] - opt_params['ci']

        if opt_params['price_diff'] > 0:
            opt_params['flag'] = 1

        else:
            opt_params['flag'] = -1

        # Starting vol shift size
        opt_params['shift'] = 0.01

        opt_params['price_diff_start'] = opt_params['price_diff']

        while abs(opt_params['price_diff']) > opt_params['epsilon']:

            # If the price difference changes sign after the vol shift,
            # reduce the decimal by one and reverse the sign
            if (np.sign(opt_params['price_diff'])
                != np.sign(opt_params['price_diff_start'])):
                opt_params['shift'] = opt_params['shift'] * -0.1

            # Calculate new vol
            opt_params['vi'] += (opt_params['shift'] * opt_params['flag'])

            # Set initial price difference
            opt_params['price_diff_start'] = opt_params['price_diff']

            # Calculate the option price with new vol
            opt_params['ci'] = cls.black_scholes_merton(
                opt_params=opt_params, sigma=opt_params['vi'])

            # Price difference after shifting vol
            opt_params['price_diff'] = opt_params['cm'] - opt_params['ci']

            # If values are diverging reverse the shift sign
            if (abs(opt_params['price_diff'])
                > abs(opt_params['price_diff_start'])):
                opt_params['shift'] = -opt_params['shift']

        result = opt_params['vi']

        return result


    @classmethod
    def implied_vol_naive_verbose(cls, opt_params):
        """
        Finds implied volatility using simple naive iteration,
        increasing precision each time the difference changes sign.

        Parameters
        ----------
        S : Float
            Stock Price. The default is 100.
        K : Float
            Strike Price. The default is 100.
        T : Float
            Time to Maturity.  The default is 0.25 (3 Months).
        r : Float
            Interest Rate. The default is 0.005 (50bps)
        q : Float
            Dividend Yield.  The default is 0.
        cm : Float
            # Option price used to solve for vol. The default is 5.
        epsilon : Float
            Degree of precision. The default is 0.0001
        option : Str
            Type of option. 'put' or 'call'. The default is 'call'.

        Returns
        -------
        result : Float
            Implied Volatility.

        """

        opt_params['vi'] = 0.2
        opt_params['ci'] = cls.black_scholes_merton(
            opt_params=opt_params, sigma=opt_params['vi'])

        opt_params['price_diff'] = opt_params['cm'] - opt_params['ci']
        if opt_params['price_diff'] > 0:
            opt_params['flag'] = 1
        else:
            opt_params['flag'] = -1
        while abs(opt_params['price_diff']) > opt_params['epsilon']:
            while opt_params['price_diff'] * opt_params['flag'] > 0:
                opt_params['ci'] = cls.black_scholes_merton(
                    opt_params=opt_params, sigma=opt_params['vi'])

                opt_params['price_diff'] = opt_params['cm'] - opt_params['ci']
                opt_params['vi'] += (0.01 * opt_params['flag'])

            while opt_params['price_diff'] * opt_params['flag'] < 0:
                opt_params['ci'] = cls.black_scholes_merton(
                    opt_params=opt_params, sigma=opt_params['vi'])

                opt_params['price_diff'] = opt_params['cm'] - opt_params['ci']
                opt_params['vi'] -= (0.001 * opt_params['flag'])

            while opt_params['price_diff'] * opt_params['flag'] > 0:
                opt_params['ci'] = cls.black_scholes_merton(
                    opt_params=opt_params, sigma=opt_params['vi'])

                opt_params['price_diff'] = opt_params['cm'] - opt_params['ci']
                opt_params['vi'] += (0.0001 * opt_params['flag'])

            while opt_params['price_diff'] * opt_params['flag'] < 0:
                opt_params['ci'] = cls.black_scholes_merton(
                    opt_params=opt_params, sigma=opt_params['vi'])

                opt_params['price_diff'] = opt_params['cm'] - opt_params['ci']
                opt_params['vi'] -= (0.00001 * opt_params['flag'])

            while opt_params['price_diff'] * opt_params['flag'] > 0:
                opt_params['ci'] = cls.black_scholes_merton(
                    opt_params=opt_params, sigma=opt_params['vi'])

                opt_params['price_diff'] = opt_params['cm'] - opt_params['ci']
                opt_params['vi'] += (0.000001 * opt_params['flag'])

            while opt_params['price_diff'] * opt_params['flag'] < 0:
                opt_params['ci'] = cls.black_scholes_merton(
                    opt_params=opt_params, sigma=opt_params['vi'])

                opt_params['price_diff'] = opt_params['cm'] - opt_params['ci']
                opt_params['vi'] -= (0.0000001 * opt_params['flag'])

        result = opt_params['vi']

        return result


    @staticmethod
    def black_scholes_merton(opt_params, sigma):
        """
        Black-Scholes-Merton Option price

        Parameters
        ----------
        S : Float
            Stock Price. The default is 100.
        K : Float
            Strike Price. The default is 100.
        T : Float
            Time to Maturity.  The default is 0.25 (3 Months).
        r : Float
            Interest Rate. The default is 0.005 (50bps)
        q : Float
            Dividend Yield.  The default is 0.
        sigma : Float
            Implied Volatility.  The default is 0.2 (20%).
        option : Str
            Type of option. 'put' or 'call'. The default is 'call'.

        Returns
        -------
        opt_price : Float
            Option Price.

        """

        opt_params['b'] = opt_params['r'] - opt_params['q']
        opt_params['carry'] = np.exp(
            (opt_params['b'] - opt_params['r']) * opt_params['T'])
        opt_params['d1'] = (
            (np.log(opt_params['S'] / opt_params['K'])
             + (opt_params['b'] + (0.5 * sigma ** 2)) * opt_params['T'])
              / (sigma * np.sqrt(opt_params['T'])))
        opt_params['d2'] = (
            (np.log(opt_params['S'] / opt_params['K'])
             + (opt_params['b'] - (0.5 * sigma ** 2)) * opt_params['T'])
              / (sigma * np.sqrt(opt_params['T'])))

        # Cumulative normal distribution function
        opt_params['Nd1'] = si.norm.cdf(opt_params['d1'], 0.0, 1.0)
        opt_params['minusNd1'] = si.norm.cdf(-opt_params['d1'], 0.0, 1.0)
        opt_params['Nd2'] = si.norm.cdf(opt_params['d2'], 0.0, 1.0)
        opt_params['minusNd2'] = si.norm.cdf(-opt_params['d2'], 0.0, 1.0)

        if opt_params['option'] == "call":
            opt_price = (
                (opt_params['S'] * opt_params['carry'] * opt_params['Nd1'])
                - (opt_params['K']
                   * np.exp(-opt_params['r'] * opt_params['T'])
                   * opt_params['Nd2']))

        elif opt_params['option'] == 'put':
            opt_price = (
                (opt_params['K']
                 * np.exp(-opt_params['r'] * opt_params['T'])
                 * opt_params['minusNd2'])
                - (opt_params['S']
                   * opt_params['carry']
                   * opt_params['minusNd1']))

        else:
            print("Please supply a value for option - 'put' or 'call'")

        return opt_price


    @staticmethod
    def black_scholes_merton_vega(opt_params, sigma):
        """
        Black-Scholes-Merton Option Vega

        Parameters
        ----------
        S : Float
            Stock Price. The default is 100.
        K : Float
            Strike Price. The default is 100.
        T : Float
            Time to Maturity.  The default is 0.25 (3 Months).
        r : Float
            Interest Rate. The default is 0.005 (50bps)
        q : Float
            Dividend Yield.  The default is 0.
        sigma : Float
            Implied Volatility.  The default is 0.2 (20%).

        Returns
        -------
        opt_vega : Float
            Option Vega.

        """

        opt_params['b'] = opt_params['r'] - opt_params['q']
        opt_params['carry'] = np.exp(
            (opt_params['b'] - opt_params['r']) * opt_params['T'])
        opt_params['d1'] = (
            (np.log(opt_params['S'] / opt_params['K'])
             + (opt_params['b'] + (0.5 * sigma ** 2)) * opt_params['T'])
              / (sigma * np.sqrt(opt_params['T'])))
        opt_params['nd1'] = (
            1 / np.sqrt(2 * np.pi)) * (np.exp(-opt_params['d1'] ** 2 * 0.5))

        opt_vega = (opt_params['S']
                    * opt_params['carry']
                    * opt_params['nd1']
                    * np.sqrt(opt_params['T']))

        return opt_vega
